import importlib

import pytest

from llm_agents_from_scratch.llms import __all__ as _llms_all


@pytest.mark.parametrize("name", _llms_all)
def test_llms_all_importable(name: str) -> None:
    """Tests that all names listed in llms __all__ are importable."""
    mod = importlib.import_module("llm_agents_from_scratch.llms")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None



"""Data Structures for LLMs."""

from enum import Enum

from pydantic import BaseModel, ConfigDict

from .tool import ToolCall


class ChatRole(str, Enum):
    """Roles for chat messages."""

    USER = "user"
    ASSISTANT = "assistant"
    SYSTEM = "system"
    TOOL = "tool"


class ChatMessage(BaseModel):
    """The chat message data model.

    Attributes:
        role: The role of the message.
        content: The content of the message.
        tool_calls: Tool calls associated with the message.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    role: ChatRole
    content: str
    tool_calls: list[ToolCall] | None = None


class CompleteResult(BaseModel):
    """The llm completion result data model.

    Attributes:
        response: The completion response provided by the LLM.
        full_response: Input prompt and completion text.
    """

    response: str
    prompt: str



from typing import Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from ollama import ChatResponse, GenerateResponse
from ollama import Message as OllamaMessage
from ollama import Tool as OllamaTool
from pydantic import BaseModel

from llm_agents_from_scratch.base.llm import BaseLLM
from llm_agents_from_scratch.base.tool import BaseTool
from llm_agents_from_scratch.data_structures import (
    ChatMessage,
    ChatRole,
    ToolCall,
    ToolCallResult,
)
from llm_agents_from_scratch.llms.ollama import OllamaLLM
from llm_agents_from_scratch.llms.ollama.utils import (
    DEFAULT_TOOL_RESPONSE_TEMPLATE,
    chat_message_to_ollama_message,
    get_tool_json_schema,
    ollama_message_to_chat_message,
    tool_call_result_to_chat_message,
    tool_to_ollama_tool,
)


def test_ollama_llm_class() -> None:
    names_of_base_classes = [b.__name__ for b in OllamaLLM.__mro__]
    assert BaseLLM.__name__ in names_of_base_classes


@patch("llm_agents_from_scratch.llms.ollama.llm.AsyncClient")
def test_init(mock_async_client_class: MagicMock) -> None:
    """Tests init of OllamaLLM."""
    mock_instance = MagicMock()
    mock_async_client_class.return_value = mock_instance
    llm = OllamaLLM(model="llama3.2")

    assert llm.model == "llama3.2"
    assert llm._client == mock_instance
    mock_async_client_class.assert_called_once()


@pytest.mark.asyncio
@patch("llm_agents_from_scratch.llms.ollama.llm.AsyncClient")
async def test_complete(mock_async_client_class: MagicMock) -> None:
    """Test complete method."""
    # arrange mocks
    mock_instance = MagicMock()
    mock_generate = AsyncMock()
    mock_generate.return_value = GenerateResponse(
        model="llama3.2",
        response="fake response",
    )
    mock_instance.generate = mock_generate
    mock_async_client_class.return_value = mock_instance

    llm = OllamaLLM(model="llama3.2")

    # act
    result = await llm.complete("fake prompt")

    # assert
    assert result.response == "fake response"
    assert result.prompt == "fake prompt"


@pytest.mark.asyncio
@patch("llm_agents_from_scratch.llms.ollama.llm.AsyncClient")
async def test_structured_output(mock_async_client_class: MagicMock) -> None:
    """Test structured_output method."""

    # Structured output type
    class Pet(BaseModel):
        animal: str
        name: str

    # arrange mocks
    mock_instance = MagicMock()
    mock_chat = AsyncMock()
    mock_chat.return_value = ChatResponse(
        model="llama3.2",
        message=OllamaMessage(
            role="assistant",
            content=Pet(animal="dog", name="spot").model_dump_json(),
        ),
    )
    mock_instance.chat = mock_chat
    mock_async_client_class.return_value = mock_instance

    llm = OllamaLLM(model="llama3.2")

    # act
    new_pet = await llm.structured_output("Generate a pet.", mdl=Pet)

    assert isinstance(new_pet, Pet)
    assert new_pet.animal == "dog"
    assert new_pet.name == "spot"
    mock_chat.assert_awaited_once_with(
        model="llama3.2",
        messages=[OllamaMessage(role="user", content="Generate a pet.")],
        format=Pet.model_json_schema(),
    )
    mock_async_client_class.assert_called_once()


@pytest.mark.asyncio
@patch("llm_agents_from_scratch.llms.ollama.llm.AsyncClient")
async def test_chat(mock_async_client_class: MagicMock) -> None:
    """Test chat method."""
    # arrange mocks
    mock_instance = MagicMock()
    mock_chat = AsyncMock()
    mock_chat.return_value = ChatResponse(
        model="llama3.2",
        message=OllamaMessage(
            role="assistant",
            content="some fake content",
            tool_calls=[
                OllamaMessage.ToolCall(
                    function=OllamaMessage.ToolCall.Function(
                        name="a_fake_tool",
                        arguments={"arg1": 1},
                    ),
                ),
            ],
        ),
    )
    mock_instance.chat = mock_chat
    mock_async_client_class.return_value = mock_instance

    llm = OllamaLLM(model="llama3.2")

    # act
    result = await llm.chat("Some new input.")

    assert result.role == "assistant"
    assert result.content == "some fake content"
    mock_chat.assert_awaited_once_with(
        model="llama3.2",
        messages=[OllamaMessage(role="user", content="Some new input.")],
        tools=None,
    )
    mock_async_client_class.assert_called_once()


@pytest.mark.asyncio
@patch("llm_agents_from_scratch.llms.ollama.llm.AsyncClient")
async def test_continue_conversation_with_tool_results(
    mock_async_client_class: MagicMock,
) -> None:
    """Test continue_conversation_with_tool_results method."""

    # arrange mocks
    mock_instance = MagicMock()
    mock_chat = AsyncMock()
    mock_chat.return_value = ChatResponse(
        model="llama3.2",
        message=OllamaMessage(
            role="assistant",
            content="Thank you for the tool call results.",
        ),
    )
    mock_instance.chat = mock_chat
    mock_async_client_class.return_value = mock_instance

    llm = OllamaLLM(model="llama3.2")

    # act
    tool_call_results = [
        ToolCallResult(
            tool_call=ToolCall(
                tool_name="a fake tool",
                arguments={"arg1": 1},
            ),
            content="Some content",
            error=False,
        ),
    ]
    new_messages = await llm.continue_conversation_with_tool_results(
        tool_call_results=tool_call_results,
        chat_messages=[],
    )

    assert len(new_messages) == len(tool_call_results) + 1
    assert new_messages[-1].role == "assistant"
    assert new_messages[-1].content == "Thank you for the tool call results."
    mock_chat.assert_awaited_once_with(
        model="llama3.2",
        messages=[
            chat_message_to_ollama_message(
                tool_call_result_to_chat_message(tool_call_results[0]),
            ),
        ],
    )
    mock_async_client_class.assert_called_once()


# test converter methods
def test_chat_message_to_ollama_message() -> None:
    """Tests conversion from ChatMessage to ~ollama.Message."""
    messages = [
        ChatMessage(
            role="system",
            content="0",
        ),
        ChatMessage(
            role="user",
            content="1",
        ),
        ChatMessage(
            role="assistant",
            content="2",
            tool_calls=[
                ToolCall(
                    tool_name="a tool",
                    arguments={
                        "arg1": "1",
                        "arg2": 2,
                    },
                ),
            ],
        ),
        ChatMessage(
            role="tool",
            content="3",
        ),
    ]

    ollama_messages = [chat_message_to_ollama_message(m) for m in messages]

    assert ollama_messages[0].content == "0"
    assert ollama_messages[0].role == "system"
    assert ollama_messages[0].tool_calls is None

    assert ollama_messages[1].content == "1"
    assert ollama_messages[1].role == "user"
    assert ollama_messages[1].tool_calls is None

    assert ollama_messages[2].content == "2"
    assert ollama_messages[2].role == "assistant"
    assert ollama_messages[2].tool_calls[0].function.name == "a tool"
    assert ollama_messages[2].tool_calls[0].function.arguments == {
        "arg1": "1",
        "arg2": 2,
    }

    assert ollama_messages[3].content == "3"
    assert ollama_messages[3].role == "tool"
    assert ollama_messages[3].tool_calls is None


def test_ollama_message_to_chat_message() -> None:
    """Tests conversion from ~ollama.Message to ChatMessage."""
    messages = [
        OllamaMessage(
            role="system",
            content="0",
        ),
        OllamaMessage(
            role="user",
            content="1",
        ),
        OllamaMessage(
            role="assistant",
            content="2",
            tool_calls=[
                OllamaMessage.ToolCall(
                    function=OllamaMessage.ToolCall.Function(
                        name="fake tool",
                        arguments={
                            "fake_param": "1",
                            "another_fake_param": "2",
                        },
                    ),
                ),
            ],
        ),
        OllamaMessage(
            role="tool",
            content="3",
        ),
    ]

    converted = [ollama_message_to_chat_message(m) for m in messages]

    assert converted[0].role == ChatRole.SYSTEM
    assert converted[0].content == "0"
    assert converted[0].tool_calls is None

    assert converted[1].role == ChatRole.USER
    assert converted[1].content == "1"
    assert converted[1].tool_calls is None

    assert converted[2].role == ChatRole.ASSISTANT
    assert converted[2].content == "2"
    assert converted[2].tool_calls[0].tool_name == "fake tool"
    assert converted[2].tool_calls[0].arguments == {
        "fake_param": "1",
        "another_fake_param": "2",
    }

    assert converted[3].role == ChatRole.TOOL
    assert converted[3].content == "3"
    assert converted[3].tool_calls is None


def test_ollama_message_to_chat_message_raises_error() -> None:
    """Test conversion to chat message raises error with invalid role."""
    with pytest.raises(RuntimeError):
        ollama_message_to_chat_message(
            OllamaMessage(
                role="invalid role",
                content="0",
            ),
        )


def test_tool_call_result_to_chat_message() -> None:
    """Test conversion of tool call result to an ChatMessage."""
    tool_call_result = ToolCallResult(
        tool_call=ToolCall(
            tool_name="a fake tool",
            arguments={"arg1": 1},
        ),
        content="Some content",
        error=False,
    )

    converted = tool_call_result_to_chat_message(tool_call_result)

    assert converted.role == "tool"
    assert converted.content == DEFAULT_TOOL_RESPONSE_TEMPLATE.format(
        tool_name="a fake tool",
        arguments={"arg1": 1},
        tool_call_result="Some content",
    )


class MyTool(BaseTool):
    @property
    def name(self) -> str:
        return "my_tool"

    @property
    def description(self) -> str:
        return "mock description"

    @property
    def parameters_json_schema(self) -> dict[str, Any]:
        return {
            "type": "object",
            "properties": {
                "param1": {
                    "type": "string",
                    "description": "param 1",
                },
                "param2": {
                    "type": "number",
                    "description": "param 2",
                },
            },
            "required": ["param1"],
        }

    def __call__(
        self,
        tool_call: ToolCall,
        *args,
        **kwargs,
    ) -> ToolCallResult:
        return ToolCallResult(
            tool_call=tool_call,
            content="fake content",
        )


def test_get_tool_json_schema() -> None:
    """Tests util for getting JSON schema of a tool."""
    # arrange
    my_tool = MyTool()

    # act
    schema = get_tool_json_schema(my_tool)

    # assert
    assert schema["type"] == "function"
    assert schema["function"]["name"] == my_tool.name
    assert schema["function"]["description"] == my_tool.description
    assert schema["function"]["parameters"] == my_tool.parameters_json_schema


def test_tool_to_ollama_tool() -> None:
    """Tests tool conversion util method."""
    # arrange
    my_tool = MyTool()

    # act
    converted_tool = tool_to_ollama_tool(my_tool)

    # arrange
    assert isinstance(converted_tool, OllamaTool)
    assert converted_tool.function.name == my_tool.name
    assert converted_tool.function.description == my_tool.description
    assert len(converted_tool.function.parameters.properties) == len(
        my_tool.parameters_json_schema["properties"],
    )



"""Base Error Class for LLM Agents From Scratch."""


class LLMAgentsFromScratchError(Exception):
    """Base error for all llm-agents-from-scratch exceptions."""

    pass


class LLMAgentsFromScratchWarning(Warning):
    """Base warning for all llm-agents-from-scratch warnings."""

    pass



from .core import LLMAgentsFromScratchError, LLMAgentsFromScratchWarning
from .task_handler import TaskHandlerError

__all__ = [
    # core
    "LLMAgentsFromScratchError",
    "LLMAgentsFromScratchWarning",
    # task handler
    "TaskHandlerError",
]



from typing import Any, Sequence

import pytest

from llm_agents_from_scratch.base.llm import BaseLLM, StructuredOutputType
from llm_agents_from_scratch.base.tool import BaseTool
from llm_agents_from_scratch.data_structures import (
    ChatMessage,
    CompleteResult,
    ToolCallResult,
)


class MockBaseLLM(BaseLLM):
    async def complete(self, prompt: str) -> CompleteResult:
        result = "mock complete"
        return CompleteResult(
            response=result,
            full_response=f"{prompt} {result}",
        )

    async def structured_output(
        self,
        prompt: str,
        mdl: type[StructuredOutputType],
        **kwargs: Any,
    ) -> StructuredOutputType:
        return mdl()

    async def chat(
        self,
        chat_messages: Sequence[ChatMessage],
        tools: Sequence[BaseTool] | None = None,
        **kwargs: Any,
    ) -> ChatMessage:
        return ChatMessage(role="assistant", content="mock chat response")

    async def continue_conversation_with_tool_results(
        self,
        tool_call_results: Sequence[ToolCallResult],
        chat_messages: Sequence[ChatMessage],
        **kwargs: Any,
    ):
        return ChatMessage(
            role="assistant",
            content="mock response to tool call result",
        )


@pytest.fixture()
def mock_llm() -> BaseLLM:
    return MockBaseLLM()



"""Ollama utils."""

from typing import Any

from ollama import Message as OllamaMessage
from ollama import Tool as OllamaTool
from typing_extensions import assert_never

from llm_agents_from_scratch.base.tool import AsyncBaseTool, BaseTool
from llm_agents_from_scratch.data_structures import (
    ChatMessage,
    ChatRole,
    ToolCall,
    ToolCallResult,
)

DEFAULT_TOOL_RESPONSE_TEMPLATE = """
The below is a tool call response for a given tool call.
<tool-call>
tool name: {tool_name}
arguments: {arguments}
</tool-call>

<result>
{tool_call_result}
</result>
"""


def ollama_message_to_chat_message(
    ollama_message: OllamaMessage,
) -> ChatMessage:
    """Convert an ~ollama.Message to ChatMessage.

    Args:
        ollama_message (Message): The ~ollama.Message to convert.

    Returns:
        ChatMessage: The converted message.
    """
    # role
    match ollama_message.role:
        case "assistant":
            role = ChatRole.ASSISTANT
        case "tool":
            role = ChatRole.TOOL
        case "user":
            role = ChatRole.USER
        case "system":
            role = ChatRole.SYSTEM
        case _:
            msg = (
                "Failed to convert ~ollama.Message due to invalid role: "
                f"`{ollama_message.role}`."
            )
            raise RuntimeError(msg)

    # convert tools
    converted_tool_calls = (
        [
            ToolCall(
                tool_name=o_tool_call.function.name,
                arguments=o_tool_call.function.arguments,
            )
            for o_tool_call in ollama_message.tool_calls
        ]
        if ollama_message.tool_calls
        else None
    )

    return ChatMessage(
        role=role,
        content=ollama_message.content,
        tool_calls=converted_tool_calls,
    )


def chat_message_to_ollama_message(chat_message: ChatMessage) -> OllamaMessage:
    """Convert a ChatMessage to an ~ollama.Message type.

    Args:
        chat_message (ChatMessage): The ChatMessage instance to convert.

    Returns:
        OllamaMessage: The converted message.
    """
    # role
    match chat_message.role:
        case ChatRole.ASSISTANT:
            role = "assistant"
        case ChatRole.TOOL:
            role = "tool"
        case ChatRole.USER:
            role = "user"
        case ChatRole.SYSTEM:
            role = "system"
        case _:  # pragma: no cover
            assert_never(chat_message.role)

    # convert tool calls
    converted_tool_calls = (
        [
            OllamaMessage.ToolCall(
                function=OllamaMessage.ToolCall.Function(
                    name=tc.tool_name,
                    arguments=tc.arguments,
                ),
            )
            for tc in chat_message.tool_calls
        ]
        if chat_message.tool_calls
        else None
    )

    return OllamaMessage(
        role=role,
        content=chat_message.content,
        tool_calls=converted_tool_calls,
    )


def tool_call_result_to_chat_message(
    tool_call_result: ToolCallResult,
) -> ChatMessage:
    """Convert a tool call result to an ~ollama.Message.

    Args:
        tool_call_result (ToolCallResult): The tool call result.

    Returns:
        ChatMessage: The converted message.
    """
    formatted_content = DEFAULT_TOOL_RESPONSE_TEMPLATE.format(
        tool_name=tool_call_result.tool_call.tool_name,
        arguments=tool_call_result.tool_call.arguments,
        tool_call_result=tool_call_result.content,
    )

    return ChatMessage(
        role=ChatRole.TOOL,
        content=formatted_content,
    )


def get_tool_json_schema(tool: BaseTool | AsyncBaseTool) -> dict[str, Any]:
    """Prepare a tool as a JSON schema.

    Args:
        tool (BaseTool | AsyncBaseTool): The tool for which to get the JSON
            schema.

    Returns:
        dict[str, Any]: The JSON schema for the tool.
    """
    return {
        "type": "function",
        "function": {
            "name": tool.name,
            "description": tool.description,
            "parameters": tool.parameters_json_schema,
        },
    }


def tool_to_ollama_tool(tool: BaseTool | AsyncBaseTool) -> OllamaTool:
    """Convert a BaseTool or AsyncBaseTool to an ~ollama.Tool type.

    Args:
        tool (BaseTool | AsyncBaseTool): The base tool to convert.

    Returns:
        ~ollama.Tool: The converted tool.
    """
    json_schema = get_tool_json_schema(tool)
    return OllamaTool.model_validate(json_schema)



from .agent import Task, TaskResult, TaskStep, TaskStepResult
from .llm import ChatMessage, ChatRole, CompleteResult
from .tool import ToolCall, ToolCallResult

__all__ = [
    # agent
    "Task",
    "TaskResult",
    "TaskStep",
    "TaskStepResult",
    # llm
    "ChatRole",
    "ChatMessage",
    "CompleteResult",
    # tool
    "ToolCall",
    "ToolCallResult",
]



import importlib

import pytest

from llm_agents_from_scratch import __all__ as _root_all


@pytest.mark.parametrize("name", _root_all)
def test_root_names_all_importable(name: str) -> None:
    """Tests that all names listed in root __all__ are importable."""
    mod = importlib.import_module("llm_agents_from_scratch")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None



"""Base LLM."""

from abc import ABC, abstractmethod
from typing import Any, Sequence, TypeVar

from pydantic import BaseModel

from llm_agents_from_scratch.data_structures import (
    ChatMessage,
    CompleteResult,
    ToolCallResult,
)

from .tool import AsyncBaseTool, BaseTool

StructuredOutputType = TypeVar("StructuredOutputType", bound=BaseModel)


class BaseLLM(ABC):
    """Base LLM Class."""

    @abstractmethod
    async def complete(self, prompt: str, **kwargs: Any) -> CompleteResult:
        """Text Complete.

        Args:
            prompt (str): The prompt the LLM should use as input.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            str: The completion of the prompt.
        """

    @abstractmethod
    async def structured_output(
        self,
        prompt: str,
        mdl: type[StructuredOutputType],
        **kwargs: Any,
    ) -> StructuredOutputType:
        """Structured output interface for returning ~pydantic.BaseModels.

        Args:
            prompt (str): The prompt to elicit the structured output response.
            mdl (type[StructuredOutputType]): The ~pydantic.BaseModel to output.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            StructuredOutputType: The structured output as the specified `mdl`
                type.
        """

    @abstractmethod
    async def chat(
        self,
        input: str,
        chat_messages: Sequence[ChatMessage] | None = None,
        tools: Sequence[BaseTool | AsyncBaseTool] | None = None,
        **kwargs: Any,
    ) -> ChatMessage:
        """Chat interface.

        Args:
            input (str): The user's current input.
            chat_messages (Sequence[ChatMessage]|None, optional): chat history.
            tools (Sequence[BaseTool]|None, optional): tools that the LLM
                can call.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            ChatMessage: The response of the LLM structured as a `ChatMessage`.
        """

    @abstractmethod
    async def continue_conversation_with_tool_results(
        self,
        tool_call_results: Sequence[ToolCallResult],
        chat_messages: Sequence[ChatMessage],
        **kwargs: Any,
    ) -> list[ChatMessage]:
        """Continue a conversation submitting tool call results.

        Args:
            tool_call_results (Sequence[ToolCallResult]):
                Tool call results.
            chat_messages (Sequence[ChatMessage]): The chat history.
                Defaults to None.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            list[ChatMessage]: The chat messages that continue the provided
                conversation history. This should include the tool call
                results as chat messages as well as the LLM's response to the
                tool call results.
        """



from llm_agents_from_scratch.base.llm import BaseLLM


def test_base_abstract_attr() -> None:
    """Tests abstract methods in base class."""
    abstract_methods = BaseLLM.__abstractmethods__

    assert "complete" in abstract_methods
    assert "chat" in abstract_methods
    assert "structured_output" in abstract_methods
    assert "continue_conversation_with_tool_results" in abstract_methods



import importlib

import pytest

from llm_agents_from_scratch.data_structures import (
    __all__ as _data_structures_all,
)


@pytest.mark.parametrize("name", _data_structures_all)
def test_data_structures_all_importable(name: str) -> None:
    """Tests all names listed in data_structures __all__ are importable."""
    mod = importlib.import_module("llm_agents_from_scratch.data_structures")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None



"""Ollama LLM integration."""

from typing import Any, Sequence

from ollama import AsyncClient

from llm_agents_from_scratch.base.llm import BaseLLM, StructuredOutputType
from llm_agents_from_scratch.base.tool import AsyncBaseTool, BaseTool
from llm_agents_from_scratch.data_structures import (
    ChatMessage,
    CompleteResult,
    ToolCallResult,
)

from .utils import (
    chat_message_to_ollama_message,
    ollama_message_to_chat_message,
    tool_call_result_to_chat_message,
    tool_to_ollama_tool,
)


class OllamaLLM(BaseLLM):
    """Ollama LLM class.

    Integration to `ollama` library for running open source models locally.
    """

    def __init__(self, model: str, *args: Any, **kwargs: Any) -> None:
        """Create an OllamaLLM instance.

        Args:
            model (str): The name of the LLM model.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.model = model
        self._client = AsyncClient()

    async def complete(self, prompt: str, **kwargs: Any) -> CompleteResult:
        """Complete a prompt with an Ollama LLM.

        Args:
            prompt (str): The prompt to complete.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            CompleteResult: The text completion result.
        """
        response = await self._client.generate(
            model=self.model,
            prompt=prompt,
            **kwargs,
        )
        return CompleteResult(
            response=response.response,
            prompt=prompt,
        )

    async def structured_output(
        self,
        prompt: str,
        mdl: type[StructuredOutputType],
        **kwargs: Any,
    ) -> StructuredOutputType:
        """Structured output interface implementation for Ollama LLM.

        Args:
            prompt (str): The prompt to elicit the structured output response.
            mdl (type[StructuredOutputType]): The ~pydantic.BaseModel to output.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            StructuredOutputType: The structured output as the specified `mdl`
                type.
        """
        o_messages = [
            chat_message_to_ollama_message(
                ChatMessage(role="user", content=prompt),
            ),
        ]
        result = await self._client.chat(
            model=self.model,
            messages=o_messages,
            format=mdl.model_json_schema(),
        )
        return mdl.model_validate_json(result.message.content)

    async def chat(
        self,
        input: str,
        chat_messages: list[ChatMessage] | None = None,
        tools: list[BaseTool | AsyncBaseTool] | None = None,
        return_history: bool = False,
        **kwargs: Any,
    ) -> ChatMessage:
        """Chat with an Ollama LLM.

        Args:
            input (str): The user's current input.
            chat_messages (list[ChatMessage] | None, optional): The chat
                history.
            tools (list[BaseTool] | None, optional): The tools available to the
                LLM.
            return_history (bool): Whether to return the update chat history.
                Defaults to False.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            ChatMessage: The chat message from the LLM.
        """
        # prepare chat history
        o_messages = [
            chat_message_to_ollama_message(
                ChatMessage(role="user", content=input),
            ),
        ]
        o_messages.extend(
            [chat_message_to_ollama_message(cm) for cm in chat_messages]
            if chat_messages
            else [],
        )

        # prepare tools
        o_tools = [tool_to_ollama_tool(t) for t in tools] if tools else None

        result = await self._client.chat(
            model=self.model,
            messages=o_messages,
            tools=o_tools,
        )

        return ollama_message_to_chat_message(result.message)

    async def continue_conversation_with_tool_results(
        self,
        tool_call_results: Sequence[ToolCallResult],
        chat_messages: Sequence[ChatMessage],
        **kwargs: Any,
    ) -> list[ChatMessage]:
        """Implements continue_conversation_with_tool_results method.

        Args:
            tool_call_results (Sequence[ToolCallResult]): The tool call results.
            chat_messages (Sequence[ChatMessage]): The chat history.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            list[ChatMessage]: The chat messages that continue the provided
                conversation history. This should include the tool call
                results as chat messages as well as the LLM's response to the
                tool call results.
        """
        # augment chat messages and convert to Ollama messages
        tool_messages = [
            tool_call_result_to_chat_message(tc) for tc in tool_call_results
        ]
        o_messages = [
            chat_message_to_ollama_message(cm) for cm in chat_messages
        ] + [chat_message_to_ollama_message(tm) for tm in tool_messages]

        # send chat request
        o_result = await self._client.chat(
            model=self.model,
            messages=o_messages,
        )

        return tool_messages + [
            ollama_message_to_chat_message(o_result.message),
        ]



import asyncio
from typing import Any, Sequence
from unittest.mock import MagicMock, patch

import pytest

from llm_agents_from_scratch.data_structures.tool import ToolCall
from llm_agents_from_scratch.tools.simple_function import (
    AsyncSimpleFunctionTool,
    SimpleFunctionTool,
    function_signature_to_json_schema,
)


def my_mock_fn_1(
    param1: int,
    param2: str = "x",
    *args: Any,
    **kwargs: Any,
) -> str:
    return f"{param1} and {param2}"


def my_mock_fn_2(
    param1: int | None,
    param2: Sequence[int],
) -> str:
    return ""


async def my_mock_fn_3(
    param1: int,
    param2: str = "x",
    *args: Any,
    **kwargs: Any,
) -> str:
    await asyncio.sleep(0.1)
    return f"{param1} and {param2}"


@pytest.mark.parametrize(
    ("func", "properties", "required"),
    [
        (
            my_mock_fn_1,
            {
                "param1": {
                    "type": "number",
                },
                "param2": {
                    "type": "string",
                },
            },
            ["param1"],
        ),
        (
            my_mock_fn_2,
            {
                "param1": {},
                "param2": {},
            },
            ["param1", "param2"],
        ),
        (
            my_mock_fn_3,
            {
                "param1": {
                    "type": "number",
                },
                "param2": {
                    "type": "string",
                },
            },
            ["param1"],
        ),
    ],
)
def test_function_as_json_schema(func, properties, required) -> None:
    schema = function_signature_to_json_schema(func)

    assert schema["type"] == "object"
    assert schema["properties"] == properties
    assert schema["required"] == required


def test_function_tool_init() -> None:
    """Tests SimpleFunctionTool initialization."""
    tool = SimpleFunctionTool(my_mock_fn_1, desc="mock desc")

    assert tool.name == "my_mock_fn_1"
    assert tool.description == "mock desc"
    assert tool.parameters_json_schema == function_signature_to_json_schema(
        my_mock_fn_1,
    )
    assert tool.func == my_mock_fn_1


@patch("llm_agents_from_scratch.tools.simple_function.validate")
def test_function_tool_call(mock_validate: MagicMock) -> None:
    """Tests a function tool call."""
    tool = SimpleFunctionTool(my_mock_fn_1, desc="mock desc")
    tool_call = ToolCall(
        tool_name="my_mock_fn_1",
        arguments={"param1": 1, "param2": "y"},
    )

    result = tool(tool_call=tool_call)

    assert result.content == "1 and y"
    mock_validate.assert_called_once_with(
        tool_call.arguments,
        schema=tool.parameters_json_schema,
    )
    assert result.error is False


def test_function_tool_call_returns_error() -> None:
    """Tests a function tool call raises error at validation of params."""
    tool = SimpleFunctionTool(my_mock_fn_1, desc="mock desc")
    tool_call = ToolCall(
        tool_name="my_mock_fn_1",
        arguments={"param1": "1", "param2": "y"},
    )

    result = tool(tool_call=tool_call)

    assert (
        "Failed to execute function call: '1' is not of type 'number'"
        in result.content
    )
    assert result.error is True


# async
def test_async_function_tool_init() -> None:
    """Tests AsyncSimpleFunctionTool initialization."""
    tool = AsyncSimpleFunctionTool(my_mock_fn_3, desc="mock desc")

    assert tool.name == "my_mock_fn_3"
    assert tool.description == "mock desc"
    assert tool.parameters_json_schema == function_signature_to_json_schema(
        my_mock_fn_3,
    )
    assert tool.func == my_mock_fn_3
    assert asyncio.iscoroutinefunction(tool.func)


@pytest.mark.asyncio
@patch("llm_agents_from_scratch.tools.simple_function.validate")
async def test_async_function_tool_call(mock_validate: MagicMock) -> None:
    """Tests a function tool call."""
    tool = AsyncSimpleFunctionTool(my_mock_fn_3, desc="mock desc")
    tool_call = ToolCall(
        tool_name="my_mock_fn_1",
        arguments={"param1": 1, "param2": "y"},
    )

    result = await tool(tool_call=tool_call)

    assert result.content == "1 and y"
    mock_validate.assert_called_once_with(
        tool_call.arguments,
        schema=tool.parameters_json_schema,
    )
    assert result.error is False


@pytest.mark.asyncio
async def test_async_function_tool_call_returns_error() -> None:
    """Tests a function tool call."""
    tool = AsyncSimpleFunctionTool(my_mock_fn_1, desc="mock desc")
    tool_call = ToolCall(
        tool_name="my_mock_fn_1",
        arguments={"param1": "1", "param2": "y"},
    )

    result = await tool(tool_call=tool_call)

    assert (
        "Failed to execute function call: '1' is not of type 'number'"
        in result.content
    )
    assert result.error is True



import asyncio
import contextlib
from unittest.mock import MagicMock, patch

import pytest
from typing_extensions import override

from llm_agents_from_scratch.base.llm import BaseLLM
from llm_agents_from_scratch.core import LLMAgent, TaskHandler
from llm_agents_from_scratch.data_structures.agent import (
    Task,
    TaskStep,
    TaskStepResult,
)


def test_init(mock_llm: BaseLLM) -> None:
    """Tests init of LLMAgent."""
    agent = LLMAgent(llm=mock_llm)

    assert len(agent.tools) == 0
    assert agent.llm == mock_llm


def test_add_tool(mock_llm: BaseLLM) -> None:
    """Tests add tool."""
    # arrange
    tool = MagicMock()
    agent = LLMAgent(llm=mock_llm)

    # act
    agent.add_tool(tool)

    # assert
    assert agent.tools == [tool]


@pytest.mark.asyncio
@patch("llm_agents_from_scratch.core.agent.TaskHandler")
async def test_run(
    mock_task_handler_class: MagicMock,
    mock_llm: BaseLLM,
) -> None:
    """Tests run method."""

    class MockTaskHandler(TaskHandler):
        @override
        async def get_next_step(self) -> TaskStep | None:
            await asyncio.sleep(0.1)
            return TaskStep(instruction="mock step", last_step=True)

        @override
        async def run_step(self, step: TaskStep) -> TaskStepResult:
            await asyncio.sleep(0.1)
            return TaskStepResult(
                task_step=step,
                content="mock result",
            )

    # arrange
    agent = LLMAgent(llm=mock_llm)
    task = Task(instruction="mock instruction")
    mock_handler = MockTaskHandler(task, agent.llm, agent.tools)
    mock_task_handler_class.return_value = mock_handler

    # act
    handler = agent.run(task)
    await handler

    # cleanup
    handler.background_task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await handler.background_task

    assert handler == mock_handler
    mock_task_handler_class.assert_called_once_with(
        task,
        agent.llm,
        agent.tools,
    )
    assert handler.result().content == "mock result"


@pytest.mark.asyncio
@patch("llm_agents_from_scratch.core.agent.TaskHandler")
async def test_run_exception(
    mock_task_handler_class: MagicMock,
    mock_llm: BaseLLM,
) -> None:
    """Tests run method with exception."""
    err = RuntimeError("mock error")

    class MockTaskHandler(TaskHandler):
        @override
        async def get_next_step(self) -> TaskStep:
            raise err

    # arrange
    agent = LLMAgent(llm=mock_llm)
    task = Task(instruction="mock instruction")
    mock_handler = MockTaskHandler(task, agent.llm, agent.tools)
    mock_task_handler_class.return_value = mock_handler

    # act
    handler = agent.run(task)
    await asyncio.sleep(0.1)  # Let it run

    assert handler == mock_handler
    mock_task_handler_class.assert_called_once_with(
        task,
        agent.llm,
        agent.tools,
    )
    assert handler.exception() == err



"""Data Structures for LLM Agent."""

from pydantic import BaseModel, Field


class Task(BaseModel):
    """Represents a single task with an instruction.

    Attributes:
        instruction: The instruction for the task.
    """

    instruction: str


class TaskStep(BaseModel):
    """Represents a step within a task and its own instruction.

    Attributes:
        instruction: The instruction for the task.
    """

    instruction: str = Field(
        description="The instruction for this step in the task.",
    )
    last_step: bool = Field(
        description=(
            "Whether or not this task step should be the final one "
            "because there is enough context to complete the overall task."
        ),
    )


class TaskStepResult(BaseModel):
    """The result of a task step execution.

    Attributes:
        task_step: The `TaskStep` that was executed.
        content: The content results of the execution.
        last_step: Whether or not the step was the final step for the parent
            Task.
    """

    task_step: TaskStep
    content: str | None


class TaskResult(BaseModel):
    """The result of the task execution.

    Attributes:
        task: The `Task` that was executed.
        content: The content results of the task execution.
        rollout: The rollout of the task execution.
        error: Whether or not the execution resulted in an error.
    """

    task: Task
    content: str
    rollout: str
    error: bool = False



"""Task Handler."""

import asyncio
from typing import Any

from llm_agents_from_scratch.base.llm import BaseLLM
from llm_agents_from_scratch.base.tool import AsyncBaseTool, BaseTool
from llm_agents_from_scratch.data_structures import (
    ChatMessage,
    ChatRole,
    Task,
    TaskStep,
    TaskStepResult,
    ToolCallResult,
)
from llm_agents_from_scratch.errors import TaskHandlerError

DEFAULT_GET_NEXT_INSTRUCTION_PROMPT = "{current_rollout}"
DEFAULT_SYSTEM_MESSAGE = "{original_instruction} {current_rollout}"
DEFAULT_USER_MESSAGE = "{instruction}"
DEFAULT_ROLLOUT_BLOCK_FROM_CHAT_MESSAGE = "{role}: {content}"


class TaskHandler(asyncio.Future):
    """Handler for processing tasks.

    Attributes:
        task: The task to execute.
        llm: The backbone LLM.
        tools_registry: The tools the LLM agent can use represented as a dict.
        rollout: The execution log of the task.
    """

    def __init__(
        self,
        task: Task,
        llm: BaseLLM,
        tools: list[BaseTool | AsyncBaseTool],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize a TaskHandler.

        Args:
            task (Task): The task to process.
            llm (BaseLLM): The backbone LLM.
            tools (list[BaseTool]): The tools the LLM can use.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.
        """
        super().__init__(*args, **kwargs)
        self.task = task
        self.llm = llm
        self.tools_registry = {t.name: t for t in tools}
        self.rollout = ""
        self._background_task: asyncio.Task | None = None
        self._lock: asyncio.Lock = asyncio.Lock()

    @property
    def background_task(self) -> asyncio.Task:
        """Get the background ~asyncio.Task for the handler."""
        if not self._background_task:
            raise TaskHandlerError(
                "No background task is running for this handler.",
            )
        return self._background_task

    @background_task.setter
    def background_task(self, asyncio_task: asyncio.Task) -> None:
        """Setter for background_task."""
        if self._background_task is not None:
            raise TaskHandlerError("A background task has already been set.")
        self._background_task = asyncio_task

    def _rollout_contribution_from_single_run_step(
        self,
        chat_history: list[ChatMessage],
    ) -> str:
        """Update rollout after a run_step execution."""
        rollout_contributions = []
        for msg in chat_history:
            # don't include system messages in rollout
            if msg.role == "system":
                continue
            rollout_contributions.append(
                DEFAULT_ROLLOUT_BLOCK_FROM_CHAT_MESSAGE.format(
                    role=msg.role.value,
                    content=msg.content,
                ),
            )
        return "\n".join(rollout_contributions)

    async def get_next_step(self) -> TaskStep:
        """Based on task progress, determine next step.

        Returns:
            TaskStep: The next step to run, if `None` then Task is done.
        """
        async with self._lock:
            rollout = self.rollout

        if rollout == "":
            return TaskStep(instruction=self.task.instruction, last_step=False)

        prompt = DEFAULT_GET_NEXT_INSTRUCTION_PROMPT.format(
            current_rollout=rollout,
        )
        try:
            task_step = await self.llm.structured_output(
                prompt=prompt,
                mdl=TaskStep,
            )
        except Exception as e:
            raise TaskHandlerError(f"Failed to get next step: {str(e)}") from e

        return task_step

    async def run_step(self, step: TaskStep) -> TaskStepResult:
        """Run next step of a given task.

        A single step is executed through a single-turn conversation that the
        LLM agent has with itself. In other words, it is both the `user`
        providing the instruction (from `get_next_step`) as well as the
        `assistant` that provides the result.

        Args:
            step (TaskStep): The step to execute.

        Returns:
            TaskStepResult: The result of the step execution.
        """
        async with self._lock:
            rollout = self.rollout

        # include rollout as context in the system message
        system_message = ChatMessage(
            role=ChatRole.SYSTEM,
            content=DEFAULT_SYSTEM_MESSAGE.format(
                original_instruction=self.task.instruction,
                current_rollout=rollout,
            ),
        )
        user_message = ChatMessage(
            role=ChatRole.USER,
            content=DEFAULT_USER_MESSAGE.format(
                instruction=step.instruction,
            ),
        )

        # start conversation
        response = await self.llm.chat(
            input=user_message.content,
            chat_messages=[system_message],
            tools=list(self.tools_registry.keys()),
        )

        chat_history = [
            system_message,
            user_message,
            response,
        ]

        # see if there are tool calls
        if response.tool_calls:
            tool_call_results = []
            for tool_call in response.tool_calls:
                if tool := self.tools_registry.get(tool_call.tool_name):
                    if isinstance(tool, AsyncBaseTool):
                        tool_call_result = await tool(tool_call=tool_call)
                    else:
                        tool_call_result = tool(tool_call=tool_call)
                else:
                    error_msg = (
                        f"Tool with name {tool_call.tool_name} doesn't exist.",
                    )
                    tool_call_result = ToolCallResult(
                        tool_call=tool_call,
                        error=True,
                        content=error_msg,
                    )
                tool_call_results.append(tool_call_result)

            # send tool call results back to llm to get result
            new_messages = (
                await self.llm.continue_conversation_with_tool_results(
                    tool_call_results=tool_call_results,
                    chat_messages=chat_history,
                )
            )

            # get final content and update chat history
            final_content = new_messages[-1].content
            chat_history += new_messages
        else:
            final_content = response.content

        # augment rollout from this turn
        async with self._lock:
            self.rollout += self._rollout_contribution_from_single_run_step(
                chat_history=chat_history,
            )

        return TaskStepResult(
            task_step=step,
            content=final_content,
        )



from llm_agents_from_scratch.base.tool import AsyncBaseTool, BaseTool


def test_base_abstract_attr() -> None:
    """Tests abstract methods in base class."""
    abstract_methods = BaseTool.__abstractmethods__

    assert "name" in abstract_methods
    assert "description" in abstract_methods
    assert "parameters_json_schema" in abstract_methods
    assert "__call__" in abstract_methods


def test_async_base_abstract_attr() -> None:
    """Tests abstract methods in async base class."""
    abstract_methods = AsyncBaseTool.__abstractmethods__

    assert "name" in abstract_methods
    assert "description" in abstract_methods
    assert "parameters_json_schema" in abstract_methods
    assert "__call__" in abstract_methods



"""Build an LLM agent from scratch."""

from llm_agents_from_scratch._version import VERSION

# Disable the F403 warning for wildcard imports
# ruff: noqa: F403, F401
from .core import *
from .core import __all__ as _core_all

__version__ = VERSION


__all__ = sorted(_core_all)  # noqa: PLE0605



"""Simple Function Tool."""

import inspect
from typing import Any, Awaitable, Callable, get_type_hints

from jsonschema import validate

from llm_agents_from_scratch.base.tool import AsyncBaseTool, BaseTool
from llm_agents_from_scratch.data_structures import ToolCall, ToolCallResult


def function_signature_to_json_schema(func: Callable) -> dict[str, Any]:
    """Turn a function signature into a JSON schema.

    Inspects the signature of the function and maps types to the appropriate
    JSON schema type.

    Args:
        func (Callable): The function for which to get the JSON schema.

    Returns:
        dict[str, Any]: The JSON schema
    """
    sig = inspect.signature(func)
    type_hints = get_type_hints(func)  # resolves forward references
    python_to_json_schema_type = {
        str: "string",
        int: "number",
        float: "number",
        dict: "object",
        list: "array",
        type(None): "null",
        bool: "boolean",
        tuple: "array",
        bytes: "string",
        set: "array",
    }

    properties = {}
    required = []
    for param in sig.parameters.values():
        # skip args and kwargs
        if param.name.startswith(("args", "kwargs")):
            continue

        # get type annotations
        annotation = type_hints.get(param.name, param.annotation)

        # get json schema
        if annotation in python_to_json_schema_type:
            this_params_json_schema = {
                "type": python_to_json_schema_type[annotation],
            }
        else:
            # fallback schema, that accepts everything
            this_params_json_schema = {}
        properties[param.name] = this_params_json_schema

        # check if param is required
        if param.default == inspect._empty:
            required.append(param.name)

    return {
        "type": "object",
        "properties": properties,
        "required": required,
    }


class SimpleFunctionTool(BaseTool):
    """Simple function calling tool.

    Turn a Python function into a tool for an LLM. Uses manual validation for
    JSON schema.
    """

    def __init__(
        self,
        func: Callable[..., Any],
        desc: str | None = None,
    ) -> None:
        """Initialize a SimpleFunctionTool.

        Args:
            func (Callable): The Python function to expose as a tool to the LLM.
            desc (str | None, optional): Description of the function.
                Defaults to None.
        """
        self.func = func
        self.desc = desc or func.__doc__ or f"Tool for {func.__name__}"

    @property
    def name(self) -> str:
        """Name of function tool."""
        return self.func.__name__

    @property
    def description(self) -> str:
        """Description of what this function tool does."""
        return self.desc

    @property
    def parameters_json_schema(self) -> dict[str, Any]:
        """JSON schema for tool parameters."""
        return function_signature_to_json_schema(self.func)

    def __call__(
        self,
        tool_call: ToolCall,
        *args: Any,
        **kwargs: Any,
    ) -> ToolCallResult:
        """Execute the function tool with a ToolCall.

        Args:
            tool_call (ToolCall): The ToolCall to execute.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            ToolCallResult: The result of the tool call execution.
        """
        try:
            # validate the arguments
            validate(tool_call.arguments, schema=self.parameters_json_schema)
            # execute the function
            res = self.func(**tool_call.arguments)
            content = str(res)
            error = False
        except Exception as e:
            content = f"Failed to execute function call: {e}"
            error = True

        return ToolCallResult(
            tool_call=tool_call,
            content=content,
            error=error,
        )


class AsyncSimpleFunctionTool(AsyncBaseTool):
    """Simple function calling tool.

    Turn a Python function into a tool for an LLM. Uses manual validation for
    JSON schema.
    """

    def __init__(
        self,
        func: Callable[..., Awaitable[Any]],
        desc: str | None = None,
    ) -> None:
        """Initialize a SimpleFunctionTool.

        Args:
            func (Callable[..., Awaitable[Any]]): Async function.
            desc (str | None, optional): Description of the function.
                Defaults to None.
        """
        self.func = func
        self.desc = desc or func.__doc__ or f"Tool for {func.__name__}"

    @property
    def name(self) -> str:
        """Name of function tool."""
        return self.func.__name__

    @property
    def description(self) -> str:
        """Description of what this function tool does."""
        return self.desc

    @property
    def parameters_json_schema(self) -> dict[str, Any]:
        """JSON schema for tool parameters."""
        return function_signature_to_json_schema(self.func)

    async def __call__(
        self,
        tool_call: ToolCall,
        *args: Any,
        **kwargs: Any,
    ) -> ToolCallResult:
        """Asynchronously execute the function tool with a ToolCall.

        Args:
            tool_call (ToolCall): The ToolCall to execute.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            ToolCallResult: The result of the tool call execution.
        """
        try:
            # validate the arguments
            validate(tool_call.arguments, schema=self.parameters_json_schema)
            # execute the function
            res = await self.func(**tool_call.arguments)
            content = str(res)
            error = False
        except Exception as e:
            content = f"Failed to execute function call: {e}"
            error = True

        return ToolCallResult(
            tool_call=tool_call,
            content=content,
            error=error,
        )



"""Base Tool."""

from abc import ABC, abstractmethod
from typing import Any

from llm_agents_from_scratch.data_structures.tool import (
    ToolCall,
    ToolCallResult,
)


class BaseTool(ABC):
    """Base Tool Class."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of tool."""

    @property
    @abstractmethod
    def description(self) -> str:
        """Description of what this tool does."""

    @property
    @abstractmethod
    def parameters_json_schema(self) -> dict[str, Any]:
        """JSON schema for tool parameters."""

    @abstractmethod
    def __call__(
        self,
        tool_call: ToolCall,
        *args: Any,
        **kwargs: Any,
    ) -> ToolCallResult:
        """Execute the tool call.

        Args:
            tool_call (ToolCall): The tool call to execute.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.


        Returns:
            ToolCallResult: The result of the tool call execution.
        """


class AsyncBaseTool(ABC):
    """Async Base Tool Class."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of tool."""

    @property
    @abstractmethod
    def description(self) -> str:
        """Description of what this tool does."""

    @property
    @abstractmethod
    def parameters_json_schema(self) -> dict[str, Any]:
        """JSON schema for tool parameters."""

    @abstractmethod
    async def __call__(
        self,
        tool_call: ToolCall,
        *args: Any,
        **kwargs: Any,
    ) -> ToolCallResult:
        """Asynchronously execute the tool call.

        Args:
            tool_call (ToolCall): The tool call to execute.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.


        Returns:
            ToolCallResult: The result of the tool call execution.
        """



import asyncio
import contextlib
from unittest.mock import AsyncMock

import pytest

from llm_agents_from_scratch.base.llm import BaseLLM
from llm_agents_from_scratch.core import TaskHandler
from llm_agents_from_scratch.core.task_handler import DEFAULT_SYSTEM_MESSAGE
from llm_agents_from_scratch.data_structures import (
    ChatMessage,
    ChatRole,
    Task,
    TaskStep,
    ToolCall,
)
from llm_agents_from_scratch.errors import TaskHandlerError
from llm_agents_from_scratch.tools.simple_function import (
    AsyncSimpleFunctionTool,
    SimpleFunctionTool,
)


def test_task_handler_init(
    mock_llm: BaseLLM,
) -> None:
    handler = TaskHandler(
        task=Task(instruction="mock instruction"),
        llm=mock_llm,
        tools=[],
    )

    assert handler.task.instruction == "mock instruction"
    assert handler.llm == mock_llm
    assert handler.tools_registry == {}


def test_task_handler_raises_error_when_getting_unset_bg_task(
    mock_llm: BaseLLM,
) -> None:
    handler = TaskHandler(
        task=Task(instruction="mock instruction"),
        llm=mock_llm,
        tools=[],
    )

    with pytest.raises(TaskHandlerError):
        handler.background_task  # noqa: B018


@pytest.mark.asyncio
async def test_task_handler_raises_error_when_setting_already_set_bg_task(
    mock_llm: BaseLLM,
) -> None:
    async def fn() -> None:
        await asyncio.sleep(0.1)

    handler = TaskHandler(
        task=Task(instruction="mock instruction"),
        llm=mock_llm,
        tools=[],
    )

    handler.background_task = asyncio.create_task(fn())
    with pytest.raises(TaskHandlerError):
        new_task = asyncio.create_task(fn())
        handler.background_task = new_task

    # cleanup
    handler.background_task.cancel()
    new_task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await handler.background_task
        await new_task


@pytest.mark.asyncio
async def test_get_next_step(mock_llm: BaseLLM) -> None:
    """Tests get next step."""

    handler = TaskHandler(
        task=Task(instruction="mock instruction"),
        llm=mock_llm,
        tools=[],
    )

    # initial task step
    initial_step = await handler.get_next_step()

    # update rollout and get next step
    expected_next_step = TaskStep(
        instruction="Some next instruction.",
        last_step=False,
    )
    magic_mock_llm = AsyncMock()
    magic_mock_llm.structured_output.return_value = expected_next_step
    handler.llm = magic_mock_llm
    handler.rollout = "some progress"
    next_step = await handler.get_next_step()

    assert initial_step.instruction == "mock instruction"
    assert initial_step.last_step is False
    assert next_step == expected_next_step


@pytest.mark.asyncio
async def test_get_next_step_raises_error(mock_llm: BaseLLM) -> None:
    """Tests get next step."""

    handler = TaskHandler(
        task=Task(instruction="mock instruction"),
        llm=mock_llm,
        tools=[],
    )

    # initial task step
    initial_step = await handler.get_next_step()

    # update rollout and get next step
    magic_mock_llm = AsyncMock()
    magic_mock_llm.structured_output.side_effect = RuntimeError("oops.")
    handler.llm = magic_mock_llm
    handler.rollout = "some progress"

    with pytest.raises(
        TaskHandlerError,
        match="Failed to get next step: oops.",
    ):
        await handler.get_next_step()

    assert initial_step.instruction == "mock instruction"
    assert initial_step.last_step is False


def test_private_rollout_contribution_from_single_run_step(
    mock_llm: BaseLLM,
) -> None:
    """Tests helper method to get rollout contribution from run step."""
    handler = TaskHandler(
        task=Task(instruction="mock instruction"),
        llm=mock_llm,
        tools=[],
    )
    chat_history = [
        ChatMessage(
            role=ChatRole.SYSTEM,
            content="a system message",
        ),
        ChatMessage(
            role=ChatRole.USER,
            content="a user message",
        ),
        ChatMessage(
            role=ChatRole.ASSISTANT,
            content="an assistant message",
            tool_calls=[
                ToolCall(
                    tool_name="a tool",
                    arguments={"tool_arg": 1},
                ),
            ],
        ),
        ChatMessage(
            role=ChatRole.TOOL,
            content="\n\ttool name: `a tool`\n\ttool result: 1+2=3.",
        ),
        ChatMessage(
            role=ChatRole.ASSISTANT,
            content="done!",
        ),
    ]

    # act
    rollout_contribution = handler._rollout_contribution_from_single_run_step(
        chat_history=chat_history,
    )

    expected_rollout_contribution = (
        "user: a user message\n"
        "assistant: an assistant message\n"
        "tool: \n\ttool name: `a tool`\n\ttool result: 1+2=3.\n"
        "assistant: done!"
    )
    assert rollout_contribution == expected_rollout_contribution


@pytest.mark.asyncio
async def test_run_step() -> None:
    """Tests run step."""

    def plus_one(arg1: int) -> int:
        return arg1 + 1

    # async simple tool
    async def plus_two(arg1: int) -> int:
        await asyncio.sleep(0.1)
        return arg1 + 2

    # arrange mocks
    mock_llm = AsyncMock()
    # initial chat response
    tool_calls = [
        ToolCall(
            tool_name="plus_one",
            arguments={"arg1": 1},
        ),
        ToolCall(
            tool_name="plus_two",
            arguments={"arg1": 1},
        ),
        # this tool doesn't exist
        ToolCall(
            tool_name="plus_three",
            arguments={"arg1": 1},
        ),
    ]
    mock_llm.chat.return_value = ChatMessage(
        role=ChatRole.ASSISTANT,
        content="Initial response.",
        tool_calls=tool_calls,
    )
    # continue conversation with tool calls
    mock_return_value = [
        # tool calls
        ChatMessage(
            role=ChatRole.TOOL,
            content="2",
        ),
        ChatMessage(
            role=ChatRole.TOOL,
            content="3",
        ),
        ChatMessage(
            role=ChatRole.TOOL,
            content="error: tool name `plus_three` doesn't exist",
        ),
        ChatMessage(
            role=ChatRole.ASSISTANT,
            content="The final response.",
        ),
    ]
    mock_llm.continue_conversation_with_tool_results.return_value = (
        mock_return_value
    )

    handler = TaskHandler(
        task=Task(instruction="mock instruction"),
        llm=mock_llm,
        tools=[
            SimpleFunctionTool(func=plus_one),
            AsyncSimpleFunctionTool(func=plus_two),
        ],
    )

    # act
    step = TaskStep(
        instruction="Some instruction.",
        last_step=False,
    )
    step_result = await handler.run_step(step)

    # assert
    mock_llm.chat.assert_awaited_once_with(
        input="Some instruction.",
        chat_messages=[
            ChatMessage(
                role=ChatRole.SYSTEM,
                content=DEFAULT_SYSTEM_MESSAGE.format(
                    original_instruction="mock instruction",
                    current_rollout="",
                ),
            ),
        ],
        tools=list(handler.tools_registry.keys()),
    )
    mock_llm.continue_conversation_with_tool_results.assert_awaited_once()
    assert step_result.task_step == step
    assert step_result.content == "The final response."


@pytest.mark.asyncio
async def test_run_step_without_tool_calls() -> None:
    """Tests run step."""

    # arrange mocks
    mock_llm = AsyncMock()
    mock_llm.chat.return_value = ChatMessage(
        role=ChatRole.ASSISTANT,
        content="Initial response.",
    )

    handler = TaskHandler(
        task=Task(instruction="mock instruction"),
        llm=mock_llm,
        tools=[],
    )

    # act
    step = TaskStep(
        instruction="Some instruction.",
        last_step=False,
    )
    step_result = await handler.run_step(step)

    # assert
    mock_llm.chat.assert_awaited_once_with(
        input="Some instruction.",
        chat_messages=[
            ChatMessage(
                role=ChatRole.SYSTEM,
                content=DEFAULT_SYSTEM_MESSAGE.format(
                    original_instruction="mock instruction",
                    current_rollout="",
                ),
            ),
        ],
        tools=list(handler.tools_registry.keys()),
    )
    mock_llm.continue_conversation_with_tool_results.assert_not_awaited()
    assert step_result.task_step == step
    assert step_result.content == "Initial response."



import asyncio
import re
from typing import Callable

import pytest
from pydantic import BaseModel, Field

from llm_agents_from_scratch.data_structures import ToolCall
from llm_agents_from_scratch.tools.pydantic_function import (
    AsyncPydanticFunction,
    AsyncPydanticFunctionTool,
    PydanticFunction,
    PydanticFunctionTool,
    _validate_pydantic_function,
)


class ParamSet1(BaseModel):
    param1: int = Field(description="A mock param 1")
    param2: str = Field(description="A mock param 2", default="x")


def my_mock_fn_1(params: ParamSet1) -> str:
    return f"{params.param1} and {params.param2}"


async def my_mock_async_fn_1(params: ParamSet1) -> str:
    return f"{params.param1} and {params.param2}"


@pytest.mark.parametrize(
    ("func", "expected_param_mdl"),
    [
        (my_mock_fn_1, ParamSet1),
        (my_mock_async_fn_1, ParamSet1),
    ],
)
def test_validate_pydantic_function(
    func: Callable,
    expected_param_mdl: type[BaseModel],
) -> None:
    param_mdl = _validate_pydantic_function(func)

    assert param_mdl == expected_param_mdl


def invalid_pydantic_fn_1(foo: str) -> bool:
    return False


def invalid_pydantic_fn_2(params: str) -> bool:
    return False


def invalid_pydantic_fn_3(params) -> bool:
    return False


@pytest.mark.parametrize(
    ("func", "msg"),
    [
        (
            invalid_pydantic_fn_1,
            "Validation of `func` failed: Missing `params` argument.",
        ),
        (
            invalid_pydantic_fn_2,
            (
                "Validation of `func` failed: <class 'str'> is not"
                " a subclass of `~pydantic.BaseModel`."
            ),
        ),
        (
            invalid_pydantic_fn_3,
            (
                "Validation of `func` failed: `params` argument must have "
                "type annotation."
            ),
        ),
    ],
)
def test_validate_pydantic_function_raises_error(
    func: Callable,
    msg: str,
) -> None:
    """Tests all the cases where validation of a PydanticFunction fails."""
    with pytest.raises(RuntimeError, match=re.escape(msg)):
        _validate_pydantic_function(func)


def test_pydantic_function_protocol() -> None:
    """Test interface for PydanticFunction protocol."""
    assert hasattr(PydanticFunction, "__call__")  # noqa: B004
    assert hasattr(PydanticFunction, "__name__")
    assert hasattr(PydanticFunction, "__doc__")


def test_init_pydantic_function_tool() -> None:
    """Tests PydanticFunctionTool initialization."""
    tool = PydanticFunctionTool(my_mock_fn_1)

    assert tool.name == "my_mock_fn_1"
    assert (
        tool.description == f"Tool for {my_mock_fn_1.__name__}"
    )  # default when None
    assert tool.parameters_json_schema == ParamSet1.model_json_schema()
    assert tool.func == my_mock_fn_1


def test_function_tool_call() -> None:
    """Tests a function tool call."""
    tool = PydanticFunctionTool(my_mock_fn_1, desc="mock desc")
    tool_call = ToolCall(
        tool_name="my_mock_fn_1",
        arguments={"param1": 1, "param2": "y"},
    )

    result = tool(tool_call=tool_call)

    assert result.content == "1 and y"
    assert result.error is False


def test_function_tool_call_returns_error() -> None:
    """Tests a function tool call raises error at validation of params."""
    tool = PydanticFunctionTool(my_mock_fn_1, desc="mock desc")
    tool_call = ToolCall(
        tool_name="my_mock_fn_1",
        arguments={"param1": "1.35", "param2": "y"},
    )

    result = tool(tool_call=tool_call)

    assert (
        "Input should be a valid integer, unable to parse string as an integer"
        in result.content
    )
    assert result.error is True


# async
def test_async_pydantic_function_protocol() -> None:
    """Test interface for PydanticFunction protocol."""
    assert hasattr(AsyncPydanticFunction, "__call__")  # noqa: B004
    assert hasattr(AsyncPydanticFunction, "__name__")
    assert hasattr(AsyncPydanticFunction, "__doc__")


def test_async_function_tool_init() -> None:
    """Tests AsyncPydanticFunctionTool initialization."""
    tool = AsyncPydanticFunctionTool(my_mock_async_fn_1, desc="mock desc")

    assert tool.name == "my_mock_async_fn_1"
    assert tool.description == "mock desc"
    assert tool.parameters_json_schema == ParamSet1.model_json_schema()
    assert tool.func == my_mock_async_fn_1
    assert asyncio.iscoroutinefunction(tool.func)


@pytest.mark.asyncio
async def test_async_function_tool_call() -> None:
    """Tests an async function tool call."""
    tool = AsyncPydanticFunctionTool(my_mock_async_fn_1, desc="mock desc")
    tool_call = ToolCall(
        tool_name="my_mock_fn_1",
        arguments={"param1": 1, "param2": "y"},
    )

    result = await tool(tool_call=tool_call)

    assert result.content == "1 and y"
    assert result.error is False


@pytest.mark.asyncio
async def test_async_function_tool_call_returns_error() -> None:
    """Tests an async function tool call returns error from validation."""
    tool = AsyncPydanticFunctionTool(my_mock_async_fn_1, desc="mock desc")
    tool_call = ToolCall(
        tool_name="my_mock_fn_1",
        arguments={"param1": "1.3156", "param2": "y"},
    )

    result = await tool(tool_call=tool_call)

    assert (
        "Input should be a valid integer, unable to parse string as an integer"
        in result.content
    )
    assert result.error is True



"""Data Structures for Tools."""

from typing import Any

from pydantic import BaseModel


class ToolCall(BaseModel):
    """Tool call.

    Attributes:
        tool_name: Name of tool to call.
        arguments: The arguments to pass to the tool execution.
    """

    tool_name: str
    arguments: dict[str, Any]


class ToolCallResult(BaseModel):
    """Result of a tool call execution.

    Attributes:
        content: The content of tool call.
        error: Whether or not the tool call yielded an error.
    """

    tool_call: ToolCall
    content: Any | None
    error: bool = False



"""Pydantic Function Tool."""

import inspect
from typing import Any, Awaitable, Callable, Protocol, get_type_hints

from pydantic import BaseModel

from llm_agents_from_scratch.base.tool import BaseTool
from llm_agents_from_scratch.data_structures import ToolCall, ToolCallResult


class PydanticFunction(Protocol):
    """PydanticFunction Protocol."""

    __name__: str
    __doc__: str | None

    def __call__(self, params: BaseModel, *args: Any, **kwargs: Any) -> Any:
        """Callable interface.

        Args:
            params (BaseModel): The function's params as a ~pydantic.BaseModel.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            Any: The result of the function call.
        """
        ...  # pragma: no cover


class AsyncPydanticFunction(Protocol):
    """Asynchronous PydanticFunction Protocol."""

    __name__: str
    __doc__: str | None

    async def __call__(
        self,
        params: BaseModel,
        *args: Any,
        **kwargs: Any,
    ) -> Awaitable[Any]:
        """Asynchronous callable interface.

        Args:
            params (BaseModel): The function's params as a ~pydantic.BaseModel.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            Awaitable[Any]: The result of the function call.
        """
        ...  # pragma: no cover


def _validate_pydantic_function(func: Callable) -> type[BaseModel]:
    """Validates func as a proper `PydanticFunction`.

    Args:
        func (Callable): The function to validate as `PydanticFunction`.

    Raises:
        RuntimeError: If validation of `func` fails.
    """
    sig = inspect.signature(func)
    type_hints = get_type_hints(func)  # resolves forward references

    if "params" not in sig.parameters:
        raise RuntimeError(
            "Validation of `func` failed: Missing `params` argument.",
        )

    if annotation := type_hints.get("params"):
        if not issubclass(annotation, BaseModel):
            msg = (
                f"Validation of `func` failed: {annotation} is not"
                " a subclass of `~pydantic.BaseModel`."
            )
            raise RuntimeError(msg)
    else:
        msg = (
            "Validation of `func` failed: `params` argument must have "
            "type annotation."
        )
        raise RuntimeError(msg)

    return annotation  # type: ignore [no-any-return]


class PydanticFunctionTool(BaseTool):
    """Pydantic function calling tool.

    Turn a Python function that takes in a ~pydantic.BaseModel params object
    into a tool for an LLM.

    Attributes:
        func: PydanticFunction to represent as a tool.
        params_mdl: The params BaseModel.
        desc: Description of the PydanticFunction.
    """

    def __init__(
        self,
        func: PydanticFunction,
        desc: str | None = None,
    ):
        """Initialize a PydanticFunctionTool.

        Args:
            func (PydanticFunction): The Pydantic function to expose as a tool
                to the LLM.
            desc (str | None, optional): Description of the function.
                Defaults to None.
        """
        self.func = func
        self.desc = desc or func.__doc__ or f"Tool for {func.__name__}"
        self.params_mdl = _validate_pydantic_function(func)

    @property
    def name(self) -> str:
        """Name of function tool."""
        return self.func.__name__

    @property
    def description(self) -> str:
        """Description of what this function tool does."""
        return self.desc

    @property
    def parameters_json_schema(self) -> dict[str, Any]:
        """JSON schema for tool parameters."""
        return self.params_mdl.model_json_schema()

    def __call__(
        self,
        tool_call: ToolCall,
        *args: Any,
        **kwargs: Any,
    ) -> ToolCallResult:
        """Execute the function tool with a ToolCall.

        Args:
            tool_call (ToolCall): The ToolCall to execute.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            ToolCallResult: The result of the tool call execution.
        """
        try:
            params = self.params_mdl.model_validate(tool_call.arguments)
            # execute the function
            res = self.func(params)
            content = str(res)
            error = False
        except Exception as e:
            content = f"Failed to execute function call: {e}"
            error = True

        return ToolCallResult(
            tool_call=tool_call,
            content=content,
            error=error,
        )


class AsyncPydanticFunctionTool(BaseTool):
    """Async Pydantic function calling tool.

    Turn an async Python function that takes in a ~pydantic.BaseModel params
    object into a tool for an LLM.

    Attributes:
        func: AsyncPydanticFunction to represent as a tool.
        params_mdl: The params BaseModel.
        desc: Description of the PydanticFunction.
    """

    def __init__(
        self,
        func: AsyncPydanticFunction,
        desc: str | None = None,
    ):
        """Initialize an AsyncPydanticFunctionTool.

        Args:
            func (AsyncPydanticFunction): The async Pydantic function to expose
                as a tool to the LLM.
            desc (str | None, optional): Description of the function.
                Defaults to None.
        """
        self.func = func
        self.desc = desc or func.__doc__ or f"Tool for {func.__name__}"
        self.params_mdl = _validate_pydantic_function(func)

    @property
    def name(self) -> str:
        """Name of function tool."""
        return self.func.__name__

    @property
    def description(self) -> str:
        """Description of what this function tool does."""
        return self.desc

    @property
    def parameters_json_schema(self) -> dict[str, Any]:
        """JSON schema for tool parameters."""
        return self.params_mdl.model_json_schema()

    async def __call__(
        self,
        tool_call: ToolCall,
        *args: Any,
        **kwargs: Any,
    ) -> ToolCallResult:
        """Execute the function tool with a ToolCall.

        Args:
            tool_call (ToolCall): The ToolCall to execute.
            *args (Any): Additional positional arguments.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            ToolCallResult: The result of the tool call execution.
        """
        try:
            params = self.params_mdl.model_validate(tool_call.arguments)
            # execute the function
            res = await self.func(params)
            content = str(res)
            error = False
        except Exception as e:
            content = f"Failed to execute function call: {e}"
            error = True

        return ToolCallResult(
            tool_call=tool_call,
            content=content,
            error=error,
        )



"""Agent Module."""

import asyncio

from typing_extensions import Self

from llm_agents_from_scratch.base.llm import BaseLLM
from llm_agents_from_scratch.base.tool import BaseTool
from llm_agents_from_scratch.data_structures import Task, TaskResult

from .task_handler import TaskHandler


class LLMAgent:
    """A simple LLM Agent Class."""

    def __init__(self, llm: BaseLLM, tools: list[BaseTool] | None = None):
        """Initialize an LLMAgent.

        Args:
            llm (BaseLLM): The backbone LLM of the LLM agent.
            tools (list[BaseTool], optional): The set of tools for the LLM
                agent. Defaults to None.

        """
        self.llm = llm
        self.tools = tools or []

    def add_tool(self, tool: BaseTool) -> Self:
        """Add a tool to the agents tool set.

        NOTE: Supports fluent style for convenience.

        Args:
            tool (BaseTool): The tool to equip the LLM agent.

        """
        self.tools = self.tools + [tool]
        return self

    def run(self, task: Task) -> TaskHandler:
        """Asynchronously run `task`."""
        task_handler = TaskHandler(task, self.llm, self.tools)

        async def _run() -> None:
            """Asynchronously process the task."""
            while not task_handler.done():
                try:
                    step = await task_handler.get_next_step()
                    step_result = await task_handler.run_step(step)
                    if step.last_step:
                        async with task_handler._lock:
                            rollout = task_handler.rollout

                        task_result = TaskResult(
                            task=task,
                            content=step_result.content,
                            rollout=rollout,
                        )
                        task_handler.set_result(task_result)
                except Exception as e:
                    task_handler.set_exception(e)

        task_handler.background_task = asyncio.create_task(_run())

        return task_handler
